#include <sstream>

#ifdef GTGEN_WITH_SOFA
	#include <fstream>

	//one of the two should be chosen
	//(both enabled works also; none leads to a disaster...)
	//
	//This one enables a code that produces initial situation for the SOFA
	//framework that than processes the created files and provide instruction
	//files for the MitoPacq to re-play the simulation accordingly.
	//#define GTGEN_WITH_SOFA_EXPORT_CONFIG
	//
	//This one instructs MitoPacq to simulate cell G2 phase according to
	//the instruction files produces with the SOFA framework.
	#define GTGEN_WITH_SOFA_PLAY

	#define SOFA_SKIP_STEP	5
#endif

#include "cell.h"
#include "toolbox/flowfields.h"
#include "toolbox/bp_lists.h"


template <class T>
void ExportGridMaskVoxelCoordinates(const i3d::Image3d<T> &mask,
											const T value,
											const char* fileName,
											i3d::Vector3d<size_t> const &step);


//
// look into cell_00_Samplephase.cpp to see full version of the
// template for an atomic cell phase function like this one
//

template <class MV, class PV>
void Cell<MV, PV>::DoG2Phase(const size_t noFrames)
{
	 REPORT("\n   *****   *****   cell ID=" << ID << ", STARTING G2-phase at timePoint=" << 
			  timePoint); 

	// ---------- local working masks ----------
	//local working copy of the cell mask - not needed
	//only nucleus
	i3d::Image3d<MV> maskNucleus;

	maskNucleus.SetResolution(scheduler.sceneRes);
	maskNucleus.SetOffset(scheduler.sceneMasks[timePoint]->GetOffset());
	//but the size of _nucleus_ mask image should contain the whole _cell_
#ifndef GTGEN_WITH_SOFA_EXPORT_CONFIG
	AllocateAndZeroNewMask(maskNucleus, scmCellBPList, noFrames, cellLookDistance);
#else
	AllocateAndZeroNewMask(maskNucleus, scmCellBPList, 2, cellLookDistance);
#endif

	//prepare working copy of cell nucleus with nucleoli
	RenderBPListIntoMask(maskNucleus,(MV)1,
	  scmNucleusBPList,scmNucleusBPCentre,scmNucleusOuterBPNumber);

#ifdef ENABLE_NUCLEOLI
	RenderBPListIntoMask(maskNucleus,(MV)2,
	  scmNucleoli1BPList,scmNucleoli1BPCentre,scmNucleoli1OuterBPNumber);
	RenderBPListIntoMask(maskNucleus,(MV)3,
	  scmNucleoli2BPList,scmNucleoli2BPCentre,scmNucleoli2OuterBPNumber);
	//now in the maskNucleus image: 0 means outside,
	//1 nucleus, 2 and 3 each nucleolus
#endif

#ifdef GTGEN_WITH_SOFA_EXPORT_CONFIG
#ifndef ENABLE_NUCLEOLI
	throw ERROR_REPORT("There are no nucleoli in the nucleus.");
#endif

	//save/export the mask image with the 3 labels
	char fn[100];
	sprintf(fn, "ID%.2u_init_nucleusMask.mhd", this->ID);
	maskNucleus.SaveImage(fn);

	//export the list of mask coordinates
	sprintf(fn, "ID%.2u_init_nucleusMask_coordinates.txt", this->ID);
	ExportMaskVoxelCoordinates(maskNucleus,(MV)0,fn);

	std::fstream BPlist;
	std::vector< i3d::Vector3d<float> >::const_iterator cBPiter;

	//export the "outer" list of nucleus boundary/countour points
	sprintf(fn, "ID%.2u_init_nucleusOuterBoundaryPoints.txt", this->ID);
	BPlist.open(fn,std::ios::out);

	cBPiter=scmNucleusBPList.begin();
	size_t counter=0;
	while ( (BPlist.good()) && (counter < scmNucleusOuterBPNumber)
			&& (cBPiter != scmNucleusBPList.end()) )
	{
		BPlist << cBPiter->x << " " << cBPiter->y << " "
				 << cBPiter->z << "\n" ;
		++cBPiter;
		++counter;
	}
	if (!BPlist.good())
	{
		BPlist.close();
		throw ERROR_REPORT("Error while writing nucleus position.");
	}
	BPlist.close();

	//export the "outer" list of the first nucleoli boundary/countour points
	sprintf(fn, "ID%.2u_init_nucleoli1OuterBoundaryPoints.txt", this->ID);
	BPlist.open(fn,std::ios::out);

	cBPiter=scmNucleoli1BPList.begin();
	counter=0;
	while ( (BPlist.good()) && (counter < scmNucleoli1OuterBPNumber)
			&& (cBPiter != scmNucleoli1BPList.end()) )
	{
		BPlist << cBPiter->x << " " << cBPiter->y << " "
				 << cBPiter->z << "\n" ;
		++cBPiter;
		++counter;
	}
	if (!BPlist.good())
	{
		BPlist.close();
		throw ERROR_REPORT("Error while writing nucleoli_1 position.");
	}
	BPlist.close();

	//export the "outer" list of the first nucleoli boundary/countour points
	sprintf(fn, "ID%.2u_init_nucleoli2OuterBoundaryPoints.txt", this->ID);
	BPlist.open(fn,std::ios::out);

	cBPiter=scmNucleoli2BPList.begin();
	counter=0;
	while ( (BPlist.good()) && (counter < scmNucleoli2OuterBPNumber)
			&& (cBPiter != scmNucleoli2BPList.end()) )
	{
		BPlist << cBPiter->x << " " << cBPiter->y << " "
				 << cBPiter->z << "\n" ;
		++cBPiter;
		++counter;
	}
	if (!BPlist.good())
	{
		BPlist.close();
		throw ERROR_REPORT("Error while writing nucleoli_2 position.");
	}
	BPlist.close();


	//export the list chromatin points
	sprintf(fn, "ID%.2u_init_chromatinPoints.txt", this->ID);
	BPlist.open(fn,std::ios::out);

	std::vector<Molecule>::const_iterator cCHRiter=chrDotList.begin();
	while ( (BPlist.good()) && (cCHRiter != chrDotList.end()) )
	{
		BPlist << cCHRiter->pos.x << " " << cCHRiter->pos.y << " "
				 << cCHRiter->pos.z << "\n" ;
		++cCHRiter;

	}
	if (!BPlist.good())
	{
		BPlist.close();
		throw ERROR_REPORT("Error while writing chromatin positions.");
	}
	BPlist.close();


	/*
	//export the grid points
	sprintf(fn, "ID%.2u_init_gridPoints.txt", this->ID);
	ExportGridMaskVoxelCoordinates(maskNucleus,(MV)1,fn,(size_t)5);
	*/


	//prepare non-rigid deformation plan for export
	scheduler.PnGetMeNonrigidHints(scmNucleusBPList,scmNucleusOuterBPNumber,scmPnHints);
	//note that the previous function makes sure: scmPnHints.GetSizeX() == scmNucleusOuterBPNumber

	sprintf(fn, "ID%.2u_init_nucleusOuterBoundaryPoints_DeformationMagnitudes.ics", this->ID);
	scmPnHints.SaveImage(fn);

	const float *myHintPtr=scmPnHints.GetFirstVoxelAddr();
	for (size_t curTime=0; curTime < scmPnHints.GetSizeY(); ++curTime)
	{
		sprintf(fn, "ID%.2u_T%03lu_nucleusOuterBoundaryPoints_DeformationMagnitudes.txt", this->ID,curTime);
		BPlist.open(fn,std::ios::out);

		float mult=1.f;
		if (curTime == scmPnNextStrongUsage)
		{
			//the code is a copy from Cell::ScmSuggestNonrigidDeformation()
			const float PnStrongInitDelay=10.f; //dalsi strong non-rigid bude prumerne v 10-tem frame
			scmPnNextStrongUsage+=(size_t)std::max(GetRandomGauss(PnStrongInitDelay,PnStrongInitDelay/4.f),1.f);

			DEBUG_REPORT("using stronger hints this time; next such event will be at " << scmPnNextStrongUsage);

			mult=cellNonRigidStrongDeformation/cellNonRigidDeformation;
		}

		for (size_t BPindex=0; BPindex < scmPnHints.GetSizeX(); ++BPindex, ++myHintPtr)
			BPlist << *myHintPtr *mult << "\n";

		BPlist.close();
	}

	//timePoint+=noFrames;
	//return;
#endif

	i3d::Image3d<float> *distanceMap = NULL;

	// ---------- helper flow fields ----------
	FlowField<float> FF;
	InitFlowField(maskNucleus,FF);

	// ---------- actions a phase needs to do before the frame generating cycle ----------
	//prepare the activity plan to spread the phase duties over the given number of frames

#ifndef GTGEN_WITH_SOFA_PLAY
	//the main cycle
	for (size_t T=0; T < noFrames; ++T) {
		REPORT("\n   *****   *****   cell ID=" << ID << ", doing G2-phase at timePoint=" << timePoint \
		       << " with sync");

		//
		// internals
		//

		//no change of cell shape here

		//just movement of dots:
		//over all chromosomes
		for (size_t i=0; i < chrCount; ++i) 
		{
			ChrMoveWithBrown(i,0.02f,maskNucleus,distanceMap,1);
			//ChrMoveWithGravity(i,0.25f,maskNucleus,distanceMap,1);
		}


		//
		// externals
		//

		//reset the current flow field
		FF.x->GetVoxelData()=0.f;
		FF.y->GetVoxelData()=0.f;
		if (FF.z) FF.z->GetVoxelData()=0.f;

		// bool collapse = (timePoint == scmPnNextStrongUsage) ? true : false;

		ScmUniversalMovement(FF);

		//
		// a footer common to all phase functions:
		//

		//move the boundary points and draw cell into sceneMasks
		MoveBPListAndCentre(scmCellBPList,scmCellBPCentre,FF);
		MoveBPListAndCentre(scmNucleusBPList,scmNucleusBPCentre,FF);

#ifdef ENABLE_NUCLEOLI
		MoveBPListAndCentre(scmNucleoli1BPList,scmNucleoli1BPCentre,FF);
		MoveBPListAndCentre(scmNucleoli2BPList,scmNucleoli2BPCentre,FF);
#endif

		ScmRemoveCellFromMask(*scheduler.sceneMasks[timePoint+1]);
		ScmRenderCellIntoMask(*scheduler.sceneMasks[timePoint+1]
#ifdef GTGEN_WITH_NUCLEIMASKS
			,scheduler.sceneNuclei[timePoint+1]
#endif
		);


		//update also the local masks
		//no need to update cell now, it is not used in the cycle

		//nucleous with nucleoli:
		maskNucleus.Clear();
		RenderBPListIntoMask(maskNucleus,(MV)1,
		  scmNucleusBPList,scmNucleusBPCentre,scmNucleusOuterBPNumber);

#ifdef ENABLE_NUCLEOLI
		RenderBPListIntoMask(maskNucleus,(MV)2,
		  scmNucleoli1BPList,scmNucleoli1BPCentre,scmNucleoli1OuterBPNumber);
		RenderBPListIntoMask(maskNucleus,(MV)3,
		  scmNucleoli2BPList,scmNucleoli2BPCentre,scmNucleoli2OuterBPNumber);
#endif

		if (distanceMap)
		{
			 delete distanceMap; 
			 distanceMap = NULL; // force recomputing distance map
		}

		//update dots in the masks
		ChrMoveByFlow(FF, maskNucleus, distanceMap,(MV)1);

		ChrUpdateStats();

		// if strong deformation happened, call the repair function
		// if (collapse)
		// call 'repair' function each RESTART_PERIOD-th frame
		if ((this->timePoint % RESTART_PERIOD) == RESTART_PERIOD-1)
		{ 
			 // smazme - begin
			 /*
			 i3d::Image3d<PV> kesmazani;
			 kesmazani.CopyMetaData(*scheduler.scenePhantoms[timePoint+1]);
			 char fn[50];
			 sprintf(fn, "ID%.2u_T%.3lu_predKorekci.ics", this->ID, this->timePoint);
			 DEBUG_REPORT("Saving: " << fn);
			 ChrRenderIntoPhantoms(kesmazani);
			 kesmazani.SaveImage(fn);
			 */
			 // smazme - end

			 /*
			 // select appropriate points 
			 std::vector<bool> overcrowded(chrCount);
			 for (size_t i=0; i<chrCount; i++)
			 {
				  overcrowded[i] = chrDensities[i] > 170.0f; //was: 160.0f;
			 }

			 ChrRepairDistrLocally(maskNucleus, overcrowded, distanceMap, MV(1));
			 */
			 ChrRepairDistrWeighted(maskNucleus, distanceMap, MV(1));
			 ChrUpdateStats();
		}

#ifdef GTGEN_DEBUG
		ChrReportStats();
		ChrReportClosests();
#endif

		//here, one may possibly add UnevenDistributionCorrection...
		//problematic chromosomes are likely those with chrDensities[]>25

		//draw the phantom image
		ChrRenderIntoPhantoms(*scheduler.scenePhantoms[timePoint+1]);
#ifdef GTGEN_WITH_TERRITORIES
		ChrRenderIntoTerritories(*scheduler.sceneTerritories[timePoint+1]);
#endif

		++timePoint;
	}

#else  //after "#ifndef GTGEN_WITH_SOFA_PLAY"

#ifndef GTGEN_WITH_SOFA_EXPORT_CONFIG
	//if these have not been already defined, define them now...
	char fn[100];
	std::fstream BPlist;
#endif

	//now, re-read and this re-init the intial position as the SOFA might
	//have adjusted the boundary point positions a little bit ->
	//we will, therefore, not recalculate the BPCentre positions now

	//but since our nucleus, nucleoli and chromatin is created randomly,
	//if the simulator is re-run, the current respective data structures
	//are not relevant to what we read from the SOFA files, so we're gonna
	//"reset" them:
	scmCellBPList.clear(); scmCellOuterBPNumber=0;
	scmNucleusBPList.clear();
	scmNucleoli1BPList.clear();
	scmNucleoli2BPList.clear();
	scmNucleusOuterBPNumber=scmNucleoli1OuterBPNumber=scmNucleoli2OuterBPNumber=0;

	//nucleus "initial" import
	sprintf(fn, "ID%.2u_T%03lu_nucleusOuterBoundaryPoints.txt", this->ID,SOFA_SKIP_STEP*timePoint);
	BPlist.open(fn,std::ios::in);
	if (! BPlist.is_open())
		throw ERROR_REPORT("Error opening new \"initial\" nucleus position.");

	while (BPlist.good())
	{
		float x,y,z;
		BPlist >> x >> y >> z;

		scmCellBPList.push_back(i3d::Vector3d<float>(x,y,z));
		scmNucleusBPList.push_back(i3d::Vector3d<float>(x,y,z));

		++scmCellOuterBPNumber;
		++scmNucleusOuterBPNumber;
	}
	BPlist.close();
	--scmCellOuterBPNumber; //as even the last unsuccessful attempt has been counted
	--scmNucleusOuterBPNumber;

	//nucleoli1 "initial" import
	sprintf(fn, "ID%.2u_T%03lu_nucleoli1OuterBoundaryPoints.txt", this->ID,SOFA_SKIP_STEP*timePoint);
	BPlist.open(fn,std::ios::in);
	if (! BPlist.is_open())
		throw ERROR_REPORT("Error opening new \"initial\" nucleoli_1 position.");

	while (BPlist.good())
	{
		float x,y,z;
		BPlist >> x >> y >> z;

		scmNucleoli1BPList.push_back(i3d::Vector3d<float>(x,y,z));
		++scmNucleoli1OuterBPNumber;
	}
	BPlist.close();
	--scmNucleoli1OuterBPNumber;

	//nucleoli2 "initial" import
	sprintf(fn, "ID%.2u_T%03lu_nucleoli2OuterBoundaryPoints.txt", this->ID,SOFA_SKIP_STEP*timePoint);
	BPlist.open(fn,std::ios::in);
	if (! BPlist.is_open())
		throw ERROR_REPORT("Error opening new \"initial\" nucleoli_2 position.");

	while (BPlist.good())
	{
		float x,y,z;
		BPlist >> x >> y >> z;

		scmNucleoli2BPList.push_back(i3d::Vector3d<float>(x,y,z));
		++scmNucleoli2OuterBPNumber;
	}
	BPlist.close();
	--scmNucleoli2OuterBPNumber;

	DEBUG_REPORT(scmNucleusOuterBPNumber << " nucleus (and cell) points were read");
	DEBUG_REPORT(scmNucleoli1OuterBPNumber << " and " << scmNucleoli2OuterBPNumber
		<< " nucleoli_1 and _2 points were read");

	//also reset chromatin:
	chrDotList.clear();
	//sprintf(fn, "ID%.2u_T%03lu_gridPoints.txt", this->ID,SOFA_SKIP_STEP*timePoint);
	sprintf(fn, "ID%.2u_T%03lu_chromatinPoints.txt", this->ID,SOFA_SKIP_STEP*timePoint);
	//sprintf(fn, "ID%.2u_init_chromatinPoints.txt", this->ID);
	BPlist.open(fn,std::ios::in);
	if (! BPlist.is_open())
		throw ERROR_REPORT("Error opening new \"initial\" chromatin positions.");

	//int gridCnt=19321;
	int gridCnt=19321000;
	while ((BPlist.good()) && (gridCnt > 0))
	{
		float x,y,z;
		BPlist >> x >> y >> z;
		chrDotList.push_back(Molecule(i3d::Vector3d<float>(x,y,z)));
		--gridCnt;
	}
	BPlist.close();
	//remove the last unsuccessful attempt
	chrDotList.pop_back();

	DEBUG_REPORT(chrDotList.size() << " chromatin dots were read");
	DEBUG_REPORT("using SKIP_STEP=" << SOFA_SKIP_STEP);

	chrDots=chrDotList.size() / chrCount;
	if (chrDots*chrCount != chrDotList.size())
		throw ERROR_REPORT("New \"initial\" chromatin positions is not OK");

	ChrUpdateStats();

	//now that the boundary points are read, adjust the local mask size
	//to accommodate for all the possible movement (note that the local mask
	//was initiated with rather tight boundary due to the SOFA needs), and
//T	AllocateAndZeroNewMask(maskNucleus, scmCellBPList, noFrames, cellLookDistance);

	//alocate the new additional helper flow field (FF)
	FlowField<float> slsFF;
	InitFlowField(maskNucleus,slsFF);

	//resize the existing FF to the new maskNucleus image size
	InitFlowField(maskNucleus,FF);
	

	//re-render the current scene mask according to the new
	//cell boundary points (which are apparently the nucleus)
	ScmRemoveCellFromMask(*scheduler.sceneMasks[timePoint]);
#ifdef GTGEN_WITH_NUCLEIMASKS
	//if enabled, one must erase the old already rendered nucleus mask
	MV* np=scheduler.sceneNuclei[timePoint]->GetFirstVoxelAddr();
	MV* const npL=np+scheduler.sceneNuclei[timePoint]->GetImageSize();
	for (; np != npL; ++np) if (*np == this->ID) *np=0;
#endif
	ScmRenderCellIntoMask(*scheduler.sceneMasks[timePoint]
#ifdef GTGEN_WITH_NUCLEIMASKS
		,scheduler.sceneNuclei[timePoint]
#endif
	);

	//draw the phantom image
	ChrRenderIntoPhantoms(*scheduler.scenePhantoms[timePoint]);
#ifdef GTGEN_WITH_TERRITORIES
	ChrRenderIntoTerritories(*scheduler.sceneTerritories[timePoint]);
#endif


	//the main cycle
	for (size_t T=0; T < noFrames; ++T) {
		REPORT("\n   *****   *****   cell ID=" << ID << ", doing G2-phase at timePoint=" << timePoint \
		       << " without sync, with SOFA");

		//
		// internals
		//

		//read the current cell shape, driven now by the SOFA:
		//nucleus import
		sprintf(fn, "ID%.2u_T%03lu_nucleusOuterBoundaryPoints.txt", this->ID,SOFA_SKIP_STEP*(timePoint+1));
		BPlist.open(fn,std::ios::in);
		//if it won't open here, it will fail on the very first reading

		std::vector< i3d::Vector3d<float> >::iterator BPiter=scmNucleusBPList.begin();
		std::vector< i3d::Vector3d<float> >::iterator CellBPiter=scmCellBPList.begin();
		size_t cnt=0;
		scmCellBPCentre=i3d::Vector3d<float>(0.f);
		while ( (BPlist.good()) && (BPiter != scmNucleusBPList.end())
				&& (cnt < scmNucleusOuterBPNumber) )
		{
			BPlist >> BPiter->x >> BPiter->y >> BPiter->z;
			CellBPiter->x = BPiter->x;
			CellBPiter->y = BPiter->y;
			CellBPiter->z = BPiter->z;
			scmCellBPCentre += *BPiter;
			++BPiter;
			++CellBPiter;
			++cnt;
		}
		if (!BPlist.good())
		{
			BPlist.close();
			throw ERROR_REPORT("Error while reading new nucleus position.");
		}
		BPlist.close();
		scmCellBPCentre /= (float)cnt;
		scmNucleusBPCentre = scmCellBPCentre;

		//nucleoli1 import
		sprintf(fn, "ID%.2u_T%03lu_nucleoli1OuterBoundaryPoints.txt", this->ID,SOFA_SKIP_STEP*(timePoint+1));
		BPlist.open(fn,std::ios::in);

		BPiter=scmNucleoli1BPList.begin();
		cnt=0;
		scmNucleoli1BPCentre=i3d::Vector3d<float>(0.f);
		while ( (BPlist.good()) && (BPiter != scmNucleoli1BPList.end())
				&& (cnt < scmNucleoli1OuterBPNumber) )
		{
			BPlist >> BPiter->x >> BPiter->y >> BPiter->z;
			scmNucleoli1BPCentre += *BPiter;
			++BPiter;
			++cnt;
		}
		if (!BPlist.good())
		{
			BPlist.close();
			throw ERROR_REPORT("Error while reading new nucleoli_1 position.");
		}
		BPlist.close();
		scmNucleoli1BPCentre /= (float)cnt;

		//nucleoli2 import
		sprintf(fn, "ID%.2u_T%03lu_nucleoli2OuterBoundaryPoints.txt", this->ID,SOFA_SKIP_STEP*(timePoint+1));
		BPlist.open(fn,std::ios::in);

		BPiter=scmNucleoli2BPList.begin();
		cnt=0;
		scmNucleoli2BPCentre=i3d::Vector3d<float>(0.f);
		while ( (BPlist.good()) && (BPiter != scmNucleoli2BPList.end())
				&& (cnt < scmNucleoli2OuterBPNumber) )
		{
			BPlist >> BPiter->x >> BPiter->y >> BPiter->z;
			scmNucleoli2BPCentre += *BPiter;
			++BPiter;
			++cnt;
		}
		if (!BPlist.good())
		{
			BPlist.close();
			throw ERROR_REPORT("Error while reading new nucleoli_2 position.");
		}
		BPlist.close();
		scmNucleoli2BPCentre /= (float)cnt;


		//just movement of dots:
		//new positions are given by the SOFA, import the list
		//sprintf(fn, "ID%.2u_T%03lu_gridPoints.txt", this->ID,SOFA_SKIP_STEP*(timePoint+1));
		sprintf(fn, "ID%.2u_T%03lu_chromatinPoints.txt", this->ID,SOFA_SKIP_STEP*(timePoint+1));
		BPlist.open(fn,std::ios::in);

		std::vector<Molecule>::iterator CHRiter=chrDotList.begin();
		while ( (BPlist.good()) && (CHRiter != chrDotList.end()) )
		{
			BPlist >> CHRiter->pos.x >> CHRiter->pos.y >> CHRiter->pos.z;
			++CHRiter;

		}
		if (!BPlist.good())
		{
			BPlist.close();
			throw ERROR_REPORT("Error while reading new chromatin positions.");
		}
		BPlist.close();

		//TODO SOFA: kontrola, ze nove pozice jsou uvnitr vygenerovane masky,
		//treba mimi-Brown s aktualni distanceMap
/*
		for (size_t i=0; i < chrCount; ++i) 
		{
			for (int j=0; j < 10; ++j)
				ChrMoveWithBrown(i,0.02f,maskNucleus,distanceMap,1);
		}
*/
		//
		// externals
		//

		//reset the current flow field
		/*
		FF.x->GetVoxelData()=0.f;
		FF.y->GetVoxelData()=0.f;
		if (FF.z) FF.z->GetVoxelData()=0.f;
		*/

		//update the cellBP to their current positions
//T		MoveBPListAndCentre(scmCellBPList,scmCellBPCentre,slsFF);

		//and suggest a next iteration
//T		ScmXYTranslationalMovement(FF);

		//remember the new changes
//T		ConcatenateFlowFields(slsFF,FF);

		//the slsFF is now able to drag all points read from the SOFA files
		//to their timeStamp+1 (the one currently being formed) frame positions


		//
		// a footer common to all phase functions:
		//

		//move the boundary points and draw cell into sceneMasks
//T		MoveBPListAndCentre(scmCellBPList,scmCellBPCentre,FF);
//T		MoveBPListAndCentre(scmNucleusBPList,scmNucleusBPCentre,slsFF);

		ScmRemoveCellFromMask(*scheduler.sceneMasks[timePoint+1]);
		ScmRenderCellIntoMask(*scheduler.sceneMasks[timePoint+1]
#ifdef GTGEN_WITH_NUCLEIMASKS
			,scheduler.sceneNuclei[timePoint+1]
#endif
		);

/*
		//nucleous with nucleoli:
		maskNucleus.Clear();
		RenderBPListIntoMask(maskNucleus,(MV)1,
		  scmNucleusBPList,scmNucleusBPCentre,scmNucleusOuterBPNumber);

#ifdef ENABLE_NUCLEOLI
//T		MoveBPListAndCentre(scmNucleoli1BPList,scmNucleoli1BPCentre,slsFF);
//T		MoveBPListAndCentre(scmNucleoli2BPList,scmNucleoli2BPCentre,slsFF);

		RenderBPListIntoMask(maskNucleus,(MV)2,
		  scmNucleoli1BPList,scmNucleoli1BPCentre,scmNucleoli1OuterBPNumber);
		RenderBPListIntoMask(maskNucleus,(MV)3,
		  scmNucleoli2BPList,scmNucleoli2BPCentre,scmNucleoli2OuterBPNumber);
#endif

		if (distanceMap)
		{
			 delete distanceMap; 
			 distanceMap = NULL; // force recomputing distance map
		}

		//update dots in the masks
//T		ChrMoveByFlow(slsFF, maskNucleus, distanceMap,(MV)1);
*/
		ChrUpdateStats();

#ifdef GTGEN_DEBUG
		ChrReportStats();
		ChrReportClosests();
#endif

		//draw the phantom image
		ChrRenderIntoPhantoms(*scheduler.scenePhantoms[timePoint+1]);
#ifdef GTGEN_WITH_TERRITORIES
		ChrRenderIntoTerritories(*scheduler.sceneTerritories[timePoint+1]);
#endif

		++timePoint;
	}

#endif  //after "#ifndef GTGEN_WITH_SOFA_PLAY"

	//cleaning here if necessary
	
	if (distanceMap)
		 delete distanceMap;

	if (centrosomeA) delete centrosomeA;
	if (centrosomeB) delete centrosomeB;

	// create centrosomes
	centrosomeA = new Dot;
	centrosomeB = new Dot;

	//render cell mask for the GenerateCentrosomes() function
	//into a new fresh only-this purpose image
	i3d::Image3d<bool> maskCell;
	maskCell.SetResolution(scheduler.sceneRes);
	maskCell.SetOffset(scheduler.sceneMasks[timePoint]->GetOffset());
	RenderBPListIntoNewMask(maskCell,true,
		scmCellBPList,scmCellBPCentre,scmCellOuterBPNumber,0, cellLookDistance);

	GenerateCentrosomes(maskCell, centrosomeA->pos, centrosomeB->pos);

	//since cell is doing non-rigid, boundary points move appart
	//we rather re-sample them again
	scmCellOuterBPNumber=
	  ResetBoundaryPoints(scmCellBPList,scmCellBPCentre,*scheduler.sceneMasks[timePoint],ID);
	//note that here we must use current timePoint, the one already finished

	//changing boundary points, we need to re-init the hints image
	scheduler.PnGetMeNonrigidHints(scmCellBPList,scmCellOuterBPNumber,scmPnHints);

	scmNucleusOuterBPNumber=
	  ResetBoundaryPoints(scmNucleusBPList,scmNucleusBPCentre,maskNucleus);

#ifdef ENABLE_NUCLEOLI
	scmNucleoli1OuterBPNumber=
	  ResetBoundaryPoints(scmNucleoli1BPList,scmNucleoli1BPCentre,maskNucleus,(MV)2);
	scmNucleoli2OuterBPNumber=
	  ResetBoundaryPoints(scmNucleoli2BPList,scmNucleoli2BPCentre,maskNucleus,(MV)3);
#endif
}


//----------------------------------------------------------------------
template <class T>
void ExportGridMaskVoxelCoordinates(const i3d::Image3d<T> &mask,
											const T value,
											const char* fileName,
											i3d::Vector3d<size_t> const &step)
{
	std::ofstream pointFile(fileName);
	if (!pointFile.is_open())
		throw ERROR_REPORT("Can't open the output file.");

	//time savers...
	const float xRes=mask.GetResolution().GetX();
	const float yRes=mask.GetResolution().GetY();
	const float zRes=mask.GetResolution().GetZ();
	const float xOff=mask.GetOffset().x;
	const float yOff=mask.GetOffset().y;
	const float zOff=mask.GetOffset().z;

	size_t cnt=0;

	for (size_t z=0; (z < mask.GetSizeZ()) && (pointFile.good()); ++z)
	for (size_t y=0; (y < mask.GetSizeY()) && (pointFile.good()); ++y)
	for (size_t x=0; (x < mask.GetSizeX()) && (pointFile.good()); ++x)
	if ((x % step.x == 0) || (y % step.y == 0))
	{
		const T m=mask.GetVoxel(x,y,z);

		if ( (m > 0) && ((value == 0) || (value == m)) )
		{
			//turn current pixel position into an absolute micron coordinate
			//of the voxel centre
			const float X=((float)x+0.5f)/xRes + xOff;
			const float Y=((float)y+0.5f)/yRes + yOff;
			const float Z=((float)z+0.5f)/zRes + zOff;

			pointFile << X << " " << Y << " " << Z << "\n";
			++cnt;
		}
	}
	if (!pointFile.good())
	{
		pointFile.close();
		throw ERROR_REPORT("Error while exporting into the file.");
	}
	pointFile.close();

	DEBUG_REPORT("Successfully exported " << cnt
					<< " voxels into " << fileName); 
}


//----------------------------------------------------------------------
// explicit instantiations to force explicit build of the functions 
// in this file because otherwise the linker couldn't find them
//----------------------------------------------------------------------
template class Cell<i3d::GRAY8,  i3d::GRAY16>;
template class Cell<i3d::GRAY16, i3d::GRAY16>;
